#include <FFGL.h>
#include <FFGLLib.h>

#include <iostream>
#include <fstream>
#include <strstream>
#include <cstdlib> // for srand, rand
#include <ctime>   // for time

#ifdef _DEBUG
// We do this dance because we don't want Python.h to pull in python*_d.lib,
// since python*_d.lib is not part of the standard Python binary distribution.
#   undef _DEBUG
#   include "Python.h"
#   define _DEBUG
#else
#   include "Python.h"
#endif

#include "NosuchUtil.h"
#include "Pyffle.h"

#ifdef NOSUCH_LOCK
#include <pthread.h>
#endif

// #define FFPARAM_BRIGHTNESS (0)

////////////////////////////////////////////////////////////////////////////////////////////////////
//  Plugin information
////////////////////////////////////////////////////////////////////////////////////////////////////

static CFFGLPluginInfo PluginInfo ( 
	Pyffle::CreateInstance,		// Create method
	"PY00",						// Plugin unique ID
	"Pyffle",					// Plugin name											
	1,							// API major version number 													
	000,						// API minor version number	
	1,							// Plugin major version number
	000,						// Plugin minor version number
	FF_EFFECT,					// Plugin type
	"Pyffle: Python FreeFrame Language Extension",	// Plugin description
	"by Tim Thompson - me@timthompson.com" // About
);

bool Pyffle::StaticInitialized = false;

void Pyffle::StaticInitialization()
{
	// Default debugging stuff
	// This currently duplicates the values in the definitions, but
	// eventually these values should come from an init file
	NosuchDebugLevel = 0;   // 0=minimal messages, 1=more, 2=extreme
	NosuchDebugToConsole = true;
	NosuchDebugToLog = true;
	NosuchDebugAutoFlush = true;
	NosuchDebugLogFile = NosuchPyfflePath("pyffle_debug.txt");

	read_config("defaults.txt");

	NosuchDebug(1,"Pyffle StaticInitialization!");
}

Pyffle::Pyffle() : CFreeFrameGLPlugin()
{
	NosuchDebug(1,"Pyffle constructor");


	_recompileFunc = NULL;
	_dopython = TRUE;
	_python_disabled = FALSE;
	_dotest = FALSE;
	_passthru = TRUE;

	initialized = false;
	gl_shutting_down = false;

	width = 1.0f;
	height = 1.0f;

	// Input properties
#ifdef FFGL_SOURCE
	SetMinInputs(0);
	SetMaxInputs(0);
#else
	SetMinInputs(1);
	SetMaxInputs(1);
#endif

#ifdef NOSUCH_LOCK
	NosuchLockInit(&python_mutex,"python");
#endif

	// Parameters
	// SetParamInfo(FFPARAM_BRIGHTNESS, "Brightness8", FF_TYPE_STANDARD, 0.5f);
	// m_brightness = 0.5f;

	disabled = false;
	disable_on_exception = false;
}

Pyffle::~Pyffle()
{
	NosuchDebug(1,"Pyffle destructor called");
	gl_shutting_down = true;
}

static PyObject*
getpythonfunc(PyObject *module, char *name)
{
	PyObject *f = PyObject_GetAttrString(module, name);
	if (!(f && PyCallable_Check(f))) {
		NosuchDebug("Unable to find python function: %s",name);
		return NULL;
	}
	return f;
}

PyObject*
Pyffle::python_lock_and_call(PyObject* func, PyObject *pArgs)
{
	lock_python();
	PyObject *msgValue = PyObject_CallObject(func, pArgs);
	unlock_python();
	return msgValue;
}

bool
Pyffle::python_recompileModule(const char *modulename)
{
	PyObject *pArgs;
	bool r = FALSE;

	if ( _recompileFunc == NULL ) {
		NosuchDebug("Hey, _recompileFunc is NULL!?");
		return FALSE;
	}

	pArgs = Py_BuildValue("(s)", modulename);
	if ( pArgs == NULL ) {
		NosuchDebug("Cannot create python arguments to recompile");
		goto getout;
	}            
	PyObject *msgValue = python_lock_and_call(_recompileFunc, pArgs);
	Py_DECREF(pArgs);
	if (msgValue == NULL) {
		NosuchDebug("Call to recompile of %s failed\n",modulename);
	} else if (msgValue == Py_None) {
		NosuchDebug("Call to recompile of %s returned None?\n",modulename);
	} else {
		char *msg = PyString_AsString(msgValue);
		r = (*msg != '\0') ? FALSE : TRUE;
		if ( r == FALSE ) {
			NosuchDebug("Call to recompile of %s failed, msg=%s\n",modulename,msg);
		}
		Py_DECREF(msgValue);
	}
getout:
	return r;
}

std::string
Pyffle::python_draw()
{
	NosuchAssert(_callBoundFunc);
	NosuchAssert(_processorDrawFunc);

	PyObject *pArgs = Py_BuildValue("(O)",_processorDrawFunc);
	if ( pArgs == NULL ) {
		return "Cannot create python arguments to _callBoundFunc";
	}            
	PyObject *msgobj = python_lock_and_call(_callBoundFunc, pArgs);
	Py_DECREF(pArgs);
	
	if (msgobj == NULL) {
		return "Call to _callBoundFunc failed";
	}
	if (msgobj == Py_None) {
		return "Call to _callBoundFunc returned None?";
	}
	std::string msg = std::string(PyString_AsString(msgobj));
	Py_DECREF(msgobj);
	if ( msg != "" ) {
		NosuchDebug("python_callprocessor returned msg = %s",msg.c_str());
	}
	return msg;
}

PyObject*
Pyffle::python_getProcessorObject(std::string btype)
{
	const char* b = btype.c_str();
	PyObject *pArgs = Py_BuildValue("(s)", b);
	if ( pArgs == NULL ) {
		NosuchDebug("Cannot create python arguments to _getProcessorFunc");
		return NULL;
	}            
	PyObject *obj = python_lock_and_call(_getProcessorFunc, pArgs);
	Py_DECREF(pArgs);
	if (obj == NULL) {
		NosuchDebug("Call to _getProcessorFunc failed");
		return NULL;
	}
	if (obj == Py_None) {
		NosuchDebug("Call to _getProcessorFunc returned None?");
		return NULL;
	}
	return obj;
}

static PyObject* nosuchmedia_debug(PyObject* self, PyObject* args)
{
    const char* name;
 
    if (!PyArg_ParseTuple(args, "s", &name))
        return NULL;
 
	NosuchDebug("(python) %s",name);
 
    Py_RETURN_NONE;
}
 
static PyObject* nosuchmedia_glVertex2f(PyObject* self, PyObject* args)
{
	float v1, v2;

    if (!PyArg_ParseTuple(args, "ff", &v1, &v2))
        return NULL;
 
	// NosuchDebug("(python) glVertex2f %f %f",v1,v2);
	glVertex2f( v1,v2);
 
    Py_RETURN_NONE;
}
 
static PyMethodDef PyffleMethods[] =
{
     {"debug", nosuchmedia_debug, METH_VARARGS, "Log debug output."},
     {"glVertex2f", nosuchmedia_glVertex2f, METH_VARARGS, "Invoke glVertext2f."},
     {NULL, NULL, 0, NULL}
};
 
bool Pyffle::python_init() {
	if ( ! Py_IsInitialized() ) {
		NosuchDebug("Pyffle(%s): initializing python",NosuchPluginName.c_str());
		Py_Initialize();
		if ( ! Py_IsInitialized() ) {
			python_disable("Unable to initialize python?");
			return FALSE;
		}
	} else {
		NosuchDebug("NOT initializing python, already running!");
	}

     (void) Py_InitModule("pyffle", PyffleMethods);

	// We want to add our directory to the sys.path
	std::string script = NosuchSnprintf(
		"import sys\n"
		"sys.path.insert(0,'%s')\n",
			NosuchForwardSlash(NosuchPyfflePath("python")).c_str()
		);
	PyRun_SimpleString(script.c_str());

	const char* pyffleutil = "pyffleutil";

	PyObject *pName = PyString_FromString(pyffleutil);
    _PyffleUtilModule = PyImport_Import(pName);
    Py_DECREF(pName);

	if ( _PyffleUtilModule == NULL) {
		python_disable("Unable to import "+NosuchPluginName+" module");
		return FALSE;
	}

	if (!(_recompileFunc = getpythonfunc(_PyffleUtilModule,"recompile"))) {
		python_disable("Can't get recompile function from pyffle module?!");
		return FALSE;
	}

	if (!(_callBoundFunc=getpythonfunc(_PyffleUtilModule,"callboundfunc"))) {
		python_disable("Unable to find callboundfunc func");
		return FALSE;
	}

	if (!(_getProcessorFunc=getpythonfunc(_PyffleUtilModule,"getprocessor"))) {
		python_disable("Unable to find getprocessor func");
		return FALSE;
	}

	NosuchDebug(1,"Trying to recompile pyffle.");
	// Note: it's always pyffle, no matter what the Plugin name is, see comment above
	if ( python_recompileModule(pyffleutil) == FALSE ) {
		python_disable("Unable to recompile pyffle module");
		return FALSE;
	}
	NosuchDebug(1,"recompile pyffle worked. changing processor to %s",NosuchPluginName.c_str());

	if ( !python_change_processor(NosuchPluginName) ) {
		python_disable(NosuchSnprintf("Unable to get change processor to %s",NosuchPluginName.c_str()));
		return FALSE;
	}

	return TRUE;
}

bool
Pyffle::python_change_processor(std::string behavename) {

	if ( !(_processorObj = python_getProcessorObject(behavename))) {
		return FALSE;
	}
	if ( !(_processorDrawFunc = getpythonfunc(_processorObj, "processOpenGL")) ) {
		return FALSE;
	}
	return TRUE;
}

void Pyffle::python_disable(std::string msg) {
	NosuchDebug(ERROR_OUTPUT,"python is being disabled!  msg=%s",msg.c_str());
	_python_disabled = TRUE;
}

bool Pyffle::python_reloadPyffleUtilModule() {

	PyObject* newmod = PyImport_ReloadModule(_PyffleUtilModule);
	if ( newmod == NULL) {
		python_disable("Unable to reload pyffle module");
		return FALSE;
	}
	_PyffleUtilModule = newmod;

	return TRUE;
}

int Pyffle::python_runfile(std::string filename) {
	std::string fullpath = NosuchPyfflePath("python") + "\\" + filename;
	std::ifstream f(fullpath.c_str(), std::ifstream::in);
	if ( ! f.is_open() ) {
		NosuchErrorOutput("Unable to open python file: %s",fullpath.c_str());
		return 1;
	}
	std::string contents;
	std::string line;
	while (!std::getline(f,line,'\n').eof()) {
		contents += (line+"\n");
	}
	f.close();
	int r = PyRun_SimpleString(contents.c_str());
	if ( r != 0 ) {
		NosuchErrorOutput("Error executing contents of: %s",fullpath.c_str());
	}
	return r;
}

static bool
istrue(std::string s)
{
	return(s == "true" || s == "True" || s == "1");
}

bool
Pyffle::read_config(std::string configname) {
	std::string fullpath = NosuchPyfflePath("config") + "\\" + configname;
	std::ifstream f(fullpath.c_str(), std::ifstream::in);
	if ( ! f.is_open() ) {
		NosuchErrorOutput("Unable to open config file: %s",fullpath.c_str());
		return false;
	}
	std::string line;
	while (!std::getline(f,line,'\n').eof()) {
		line.erase(line.find_last_not_of(" \n\r\t")+1); // trim the end
		std::vector<std::string> words = NosuchSplitOnAnyChar(line," =");
		if ( words.size()==0 || words[0].find("#")==0 ) {
			continue;
		}
		if ( words.size() < 2 ) {
			NosuchDebug(ERROR_OUTPUT,"Error in %s: not enough words on line - %s",
				fullpath.c_str(),line.c_str());
			continue;
		}
		if ( words.size() > 2 ) {
			NosuchDebug(ERROR_OUTPUT,"Warning in %s: too many words on line - %s",
				fullpath.c_str(),line.c_str());
		}
		if ( words[0] == "DebugLevel" ) {
			NosuchDebugLevel = atoi(words[1].c_str());
		} else if ( words[0] == "DebugToConsole" ) {
			NosuchDebugToConsole = istrue(words[1]);
		} else if ( words[0] == "DebugToLog" ) {
			NosuchDebugToLog = istrue(words[1]);
		} else if ( words[0] == "DebugAutoFlush" ) {
			NosuchDebugAutoFlush = istrue(words[1]);
		} else if ( words[0] == "DebugLogFile" ) {
			size_t pos = line.find("=") + 1;
			while ( line.at(pos) == ' ' ) {
				pos++;
			}
			NosuchDebugLogFile = NosuchPyfflePath(line.substr(pos));
		}
	}
	return true;
}

void Pyffle::_initialize() {
	if ( _dopython ) {
		if ( ! python_init() ) {
			NosuchDebug("python_init failed!");
		} else {
			NosuchDebug("python_init succeeded!");
		}
	}
}

bool Pyffle::initStuff() {

	NosuchDebug(2,"initStuff starts");

	// test_stuff();

	bool r = true;
	try {
		// static initializations
		Pyffle::_initialize();
	} catch (NosuchException& e) {
		NosuchDebug("NosuchException: %s",e.message());
		r = false;
	} catch (...) {
		// Does this really work?  Not sure
		NosuchDebug("Some other kind of exception occured!?");
		r = false;
	}
	NosuchDebug(2,"initStuff returns %s\n",r?"true":"false");
	return r;
}

DWORD Pyffle::ProcessOpenGL(ProcessOpenGLStruct *pGL)
{
	if ( gl_shutting_down ) {
		return FF_SUCCESS;
	}
	if ( disabled ) {
		return FF_SUCCESS;
	}

	if ( ! initialized ) {
		if ( ! initStuff() ) {
			NosuchDebug("initStuff failed, disabling plugin!");
			disabled = true;
			return FF_FAIL;
		}
		initialized = true;
	}

#ifdef FRAMELOOPINGTEST
	static int framenum = 0;
	static bool framelooping = FALSE;
#endif

	if ( _passthru ) {
		if (pGL->numInputTextures<1)
			return FF_FAIL;

		if (pGL->inputTextures[0]==NULL)
			return FF_FAIL;
  
		FFGLTextureStruct &Texture = *(pGL->inputTextures[0]);

		//bind the texture handle
		glBindTexture(GL_TEXTURE_2D, Texture.Handle);
	}

#ifdef DOITINPYTHON
	glDisable(GL_TEXTURE_2D); 

	glEnable(GL_BLEND); 
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glLineWidth((GLfloat)3.0f);
#endif

	bool gotexception = false;
	try {
		CATCH_NULL_POINTERS;

		NosuchDebug(2,"ProcessOpenGL");
		if ( _dopython && ! _python_disabled ) {
			python_draw();
		}
	} catch (NosuchException& e ) {
		NosuchDebug("NosuchException in python_draw : %s",e.message());
		gotexception = true;
	} catch (...) {
		NosuchDebug("UNKNOWN Exception in python_draw!");
		gotexception = true;
	}

	if ( gotexception && disable_on_exception ) {
		NosuchDebug("DISABLING Pyffle due to exception!!!!!");
		disabled = true;
	}

	glDisable(GL_BLEND); 
	glEnable(GL_TEXTURE_2D); 

#ifdef FRAMELOOPINGTEST
	int w = Texture.Width;
	int h = Texture.Height;
#define NFRAMES 300
	static GLubyte* pixelArray[NFRAMES];
	if ( framelooping ) {
		glRasterPos2i(-1,-1);
		glDrawPixels(w,h,GL_RGB,GL_UNSIGNED_BYTE,pixelArray[framenum]);
		framenum = (framenum+1)%NFRAMES;
	} else {
		if ( framenum < NFRAMES ) {
			pixelArray[framenum] = new GLubyte[w*h*3];
			glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,pixelArray[framenum]);
			framenum++;
		} else {
			framelooping = TRUE;
			framenum = 0;
		}
	}
#endif

	//disable texturemapping
	glDisable(GL_TEXTURE_2D);
	
	//restore default color
	glColor4f(1.f,1.f,1.f,1.f);
	
	return FF_SUCCESS;
}

void Pyffle::lock_python() {
	// We don't actually need this, right now, since FreeFrame plugins should never
	// be running simultaneously.
#ifdef NOSUCH_LOCK
	NosuchLock(&python_mutex,"python");
#endif
}

void Pyffle::unlock_python() {
#ifdef NOSUCH_LOCK
	NosuchUnlock(&python_mutex,"python");
#endif
}

DWORD Pyffle::GetParameter(DWORD dwIndex)
{
	return FF_FAIL;  // no parameters
}

DWORD Pyffle::SetParameter(const SetParameterStruct* pParam)
{
	return FF_FAIL;  // no parameters
}
std::string
NosuchForwardSlash(std::string filepath) {
	size_t i;
	while ( (i=filepath.find("\\")) != filepath.npos ) {
		filepath.replace(i,1,"/");
	}
	return filepath;
}

std::string NosuchPyffleDir = "";
std::string NosuchPluginName = "";

std::string
NosuchPyfflePath(std::string filepath)
{
	return NosuchPyffleDir + "\\" + filepath;
}

bool
pyffle_setdll(std::string dllpath)
{
	size_t lastslash = dllpath.find_last_of("/\\");
	size_t lastunder = dllpath.find_last_of("_");
	size_t lastdot = dllpath.find_last_of(".");
	std::string suffix = (lastdot==dllpath.npos?"":dllpath.substr(lastdot));

	if ( suffix != ".dll"
		|| lastslash == dllpath.npos
		|| lastunder == dllpath.npos
		|| lastdot == dllpath.npos ) {

		NosuchDebug("Hey! dll name (%s) isn't of the form */Pyffle_Name.dll, PLUGIN IS NOW DISABLED!",dllpath.c_str());
		return FALSE;
	}

	std::string dir = dllpath.substr(0,lastslash);
	std::string prefix = dllpath.substr(lastslash+1,lastdot-lastslash-1);
	if ( prefix.substr(0,7) != "Pyffle_") {
		NosuchDebug("Hey! plugin name name (%s) isn't of the form */Pyffle_Name.dll, PLUGIN IS NOW DISABLED!",dllpath.c_str());
		return FALSE;
	}

	NosuchPluginName = prefix.substr(7);
	size_t i = NosuchPluginName.find("_debug");
	if ( i > 0 ) {
		NosuchPluginName = NosuchPluginName.substr(0,i);
	}
	NosuchDebugPrefix = "Pyffle_"+NosuchPluginName+": ";
	NosuchPyffleDir = dir + "\\Pyffle";
	NosuchDebug("Setting NosuchPyffleDir = %s",NosuchPyffleDir.c_str());

	struct _stat statbuff;
	int e = _stat(NosuchPyffleDir.c_str(),&statbuff);
	if ( ! (e == 0 && (statbuff.st_mode | _S_IFDIR) != 0) ) {
		NosuchDebug("Hey! No directory %s!?",NosuchPyffleDir.c_str());
		return FALSE;
	}

	char id[5];
	// Compute a hash of the plugin name and use two 4-bit values
	// from it to produce the last 2 characters of the unique ID.
	// It's possible there will be a collision.
	int hash = 0;
	for ( const char* p = NosuchPluginName.c_str(); *p!='\0'; p++ ) {
		hash += *p;
	}
	id[0] = 'P';
	id[1] = 'Y';
	id[2] = 'A' + (hash & 0xf);
	id[3] = 'A' + ((hash >> 4) & 0xf);
	id[4] = '\0';
	PluginInfo.SetPluginIdAndName(id,("Pyffle_"+NosuchPluginName).c_str());
	return TRUE;
}